@@ -15,6 +15,7 @@ object
   method add : op_lst -> unit
   method delete : op_lst -> unit
   method replace : op_lst -> unit
+  method modify : (Ocamldap.mod_op * string * string list) list -> unit
   method attributes : string list
   method exists : string -> bool
   method get_value : string -> string list
@@ -94,6 +95,16 @@ object (self)
     in
       do_replace x; self#push_change `REPLACE x
 
+  method modify (x: (Ocamldap.mod_op * string * string list) list) =
+    let rec do_modify x =
+      match x with
+	  [] -> ()
+	| (`ADD, attr, values) :: t -> self#add [(attr, values)];do_modify t
+	| (`DELETE, attr, values) :: t -> self#delete [(attr, values)];do_modify t
+	| (`REPLACE, attr, values) :: t -> self#replace [(attr, values)];do_modify t
+    in
+      do_modify x
+
   method attributes = 
     let keys hash = 
       let cur = ref [] in
@@ -119,8 +130,49 @@ end
 (********************************************************************************)
 (********************************************************************************)
 (********************************************************************************)
+let to_entry ent =
+  let rec add_attrs attrs entry =
+    match attrs with
+	{attr_name = name; attr_values = values} :: tail ->
+	  entry#add [(name, (Array.to_list values))]; add_attrs tail entry
+      | [] -> entry#set_changetype MODIFY; entry
+  in
+    match ent with
+	{entry_dn = dn; entry_attrs = attrs} -> 
+	  let entry = new ldapentry in
+	    entry#set_dn dn; add_attrs attrs entry	      
+
+let iter (f: ldapentry -> unit) res =   
+  try while true
+  do
+    f (res ());
+  done
+  with
+      LDAP_Failure `SUCCESS 
+    | LDAP_Failure `LOCAL_ERROR -> ()
+      
+let rev_map (f: ldapentry -> 'a) res =
+  let lst = ref [] in
+    (try while true
+     do
+       lst := (f (res ())) :: !lst
+     done
+     with 
+	 LDAP_Failure `SUCCESS
+       | LDAP_Failure `LOCAL_ERROR -> ());
+    !lst
+
+let map (f: ldapentry -> 'a) res = List.rev (rev_map f res)
+let fold f v res =
+  let rec apply f v =
+      try let e = res () in
+	apply f (f e v)
+      with LDAP_Failure `SUCCESS
+	| LDAP_Failure `LOCAL_ERROR -> v
+    in
+      apply f v
+
 (* a connection to an ldap server *)
-let finalise_ldapcon con = con#unbind
 class ldapcon ?(version = 3) ?(async = false) ?(port = 389) host = 
 object (self)
   val hst = host
@@ -131,23 +183,30 @@ object (self)
   val mutable pwd = ""
   val mutable mth = `SIMPLE
   val mutable bound = true
+  val mutable reconnect_successful = true
   val mutable con = init ~version: version ~port: port host
 
-  initializer Gc.finalise finalise_ldapcon self
   method private reconnect =
-    unbind con;
+    if bound then unbind con;
+    bound <- false;
+    reconnect_successful <- false;
     con <- init ~version: ver ~port: prt hst;
-    self#bind bdn ~cred: pwd ~meth: mth
+    bound <- true;
+    bind_s ~who: bdn ~cred: pwd ~auth_method: mth con;
+    reconnect_successful <- true;
 
   method unbind = if bound then (unbind con;bound <- false)
 
   method update_entry (e:ldapentry) = 
+    if not (reconnect_successful && bound) then self#reconnect;
     try self#modify e#dn (List.rev e#changes); e#flush_changes
     with LDAP_Failure(`SERVER_DOWN) -> 
       self#reconnect;self#update_entry e
 		      
   method bind ?(cred = "") ?(meth = `SIMPLE) dn =
+    if not bound then (con <- init ~version: ver ~port: prt hst;bound <- true);
     bind_s ~who: dn ~cred: cred ~auth_method: meth con;
+    reconnect_successful <- true;
     bdn <- dn; pwd <- cred; mth <- meth
 
   method add (entry: ldapentry) = 
@@ -157,21 +216,25 @@ object (self)
 	| attr :: tail ->
 	    (`ADD, attr, entry#get_value attr) :: mkattrs entry tail
     in
+      if not (reconnect_successful && bound) then self#reconnect;
       try add_s con entry#dn (mkattrs entry entry#attributes)
       with LDAP_Failure(`SERVER_DOWN) ->
 	self#reconnect;self#add entry
 	
   method delete dn =
+    if not (reconnect_successful && bound) then self#reconnect;
     try delete_s con dn
     with LDAP_Failure(`SERVER_DOWN) ->
       self#reconnect;self#delete dn
 
   method modify dn mods = 
+    if not (reconnect_successful && bound) then self#reconnect;
     try modify_s con dn mods
     with LDAP_Failure(`SERVER_DOWN) ->
       self#reconnect;self#modify dn mods
 
   method modrdn dn ?(deleteoldrdn = true) newrdn =
+    if not (reconnect_successful && bound) then self#reconnect;
     try modrdn2_s con ~dn:dn ~newdn:newrdn ~deleteoldrdn:deleteoldrdn
     with LDAP_Failure(`SERVER_DOWN) ->
       self#reconnect;self#modrdn dn ~deleteoldrdn:deleteoldrdn newrdn
@@ -182,31 +245,43 @@ object (self)
     ?(attrsonly = false)
     ?(base = "")
     filter =
-    let to_entry (ent: entry) =
-      let rec add_attrs attrs entry =
-	match attrs with
-	    {attr_name = name; attr_values = values} :: tail ->
-	      entry#add [(name, (Array.to_list values))]; add_attrs tail entry
- 	  | [] -> entry#set_changetype MODIFY; entry
-      in
-	match ent with
-	    {entry_dn = dn; entry_attrs = attrs} -> 
-	      let entry = new ldapentry in
-		entry#set_dn dn; add_attrs (Array.to_list attrs) entry
-    in
-      try 
-	List.rev_map to_entry (search_s 
-				 ~scope: scope
-				 ~base: base
-				 ~attrs: attrs
-				 ~attrsonly: attrsonly
-				 con
-				 filter)
-      with LDAP_Failure(`SERVER_DOWN) ->
-	self#reconnect;self#search ~scope: scope ~attrs: attrs
-	  ~attrsonly: attrsonly ~base: base filter
+    if not (reconnect_successful && bound) then self#reconnect;
+    try 
+      List.rev_map to_entry 
+	(search_s 
+	   ~scope: scope
+	   ~base: base
+	   ~attrs: attrs
+	   ~attrsonly: attrsonly
+	   con
+	   filter)
+    with LDAP_Failure(`SERVER_DOWN) ->
+      self#reconnect;self#search ~scope: scope ~attrs: attrs
+	~attrsonly: attrsonly ~base: base filter
+
+  method search_a
+    ?(scope = `SUBTREE)
+    ?(attrs = [])
+    ?(attrsonly = false)
+    ?(base = "")
+    filter =
+      let fetch_result con msgid () = to_entry (get_search_entry con msgid) in
+	if not (reconnect_successful && bound) then self#reconnect;
+	try 
+	  fetch_result con
+	    (search
+	       ~scope: scope
+	       ~base: base
+	       ~attrs: attrs
+	       ~attrsonly: attrsonly
+	       con
+	       filter)
+	with LDAP_Failure(`SERVER_DOWN) ->
+	  self#reconnect;self#search_a ~scope: scope ~attrs: attrs
+	    ~attrsonly: attrsonly ~base: base filter
 
   method schema = 
+    if not (reconnect_successful && bound) then self#reconnect;
     try 
       if ver = 3 then
 	let schema_base = (match (self#search 
@@ -229,7 +304,8 @@ object (self)
     with LDAP_Failure(`SERVER_DOWN) -> self#reconnect;self#schema
  
   method rawschema = 
-    try 
+    if not (reconnect_successful && bound) then self#reconnect;
+    try       
       if ver = 3 then
 	let schema_base = (match (self#search 
 				    ~base: "" 
@@ -655,7 +731,16 @@ object (self)
 (* evaluates the set of missing attributes to see if any of
    them can be generated, if so, it adds them to be generated *)
   method private resolve_missing =
+    (* computes the set of generateable attributes *)
     let generate_togenerate generators missing togenerate =
+      (* generators have dependancies. Some of the dependancies can
+	 also be generated. We can generate a dependancy if the following
+	 conditions are met. 
+	 1. The dependancy is in the generators hash (it has a generation function)
+	 2. The dependancy is allowed by the schema (it is either a must or may of
+	 an objectclass currently on the object)
+	 3. The dependancy is not already present (if it is present already then it
+	 has already been satisfied, and there is no need to generate it) *)
       let find_generatable_dep generators generator =
 	(List.rev_map
 	   (fun e -> attrToOid schema (Lcstring.of_string e))
@@ -667,14 +752,19 @@ object (self)
 			     (setOfList self#list_present Setstr.empty)))) then
 		   true
 		 else false)
-	      (Hashtbl.find generators generator).required))
+	      (List.filter (* we can only add it if it is allowed by the schema *)
+		 (fun attr -> super#is_allowed attr)
+		 (Hashtbl.find generators generator).required)))
       in
+	(* collect a flat list of all generatable dependancies *)
       let rec find_generatable_deps generators genlst =
 	(List.flatten
 	   (List.rev_map
 	      (find_generatable_dep generators)
 	      genlst))
       in
+	(* the set we are currently generating, union the set of missing attributes which we
+	   can generate. *)
       let generateing = (List.filter
 			   (fun gen -> 
 			      if (Hashtbl.mem generators (lowercase (oidToAttr schema gen))) then
@@ -684,6 +774,10 @@ object (self)
 			      missing
 			      (Setstr.elements togenerate)))
       in
+	(* the total set of generatable at any point in time is. The set
+	   we are already generating, unioned with any generatable dependancies, unioned
+	   with the set of missing attributes (required by the schema) which can be generated. 
+	   Note, the last union is done in the generateing expression above. *)
 	setOfList
 	  (List.rev_append generateing (find_generatable_deps
 					  generators
@@ -725,6 +819,7 @@ object (self)
 	      (Setstr.union 
 		 toGenerate 
 		 (setOfList super#list_present Setstr.empty))))
+
   method attributes =
     (List.rev_map (oidToAttr schema)
        (Setstr.elements
@@ -734,6 +829,7 @@ object (self)
 		   (fun a -> attrToOid schema (Lcstring.of_string a))
 		   super#attributes)
 		Setstr.empty))))
+
   method is_missing x = (not (Setstr.mem
 				(attrToOid schema (Lcstring.of_string x)) 
 				toGenerate)) 
@@ -860,8 +956,30 @@ object (self)
 		    | _ -> super#delete [(attr, [])])
 	     with Not_found -> ()))
 	 service.generate_attrs)	     	     
+
+  method service_exists service =
+    let service = (try (Hashtbl.find services service) 
+		   with Not_found -> raise (No_service service))
+    in
+      match self#adapt_service service with
+	  {svc_name=s;
+	   static_attrs=[];
+	   generate_attrs=[];
+	   depends=d} -> (match d with
+			      [] -> true
+			    | d  -> List.for_all self#service_exists d)
+	| _ -> false
+
+  method services_present =
+    Hashtbl.fold
+      (fun k v l -> 
+	 if self#service_exists v.svc_name then
+	   v.svc_name :: l
+	 else l)
+      services []
       
   method of_entry e = super#of_entry e;self#resolve_missing
+
   method add_generate x = 
     (if (Hashtbl.mem generators (lowercase x)) then
        toGenerate <- Setstr.add (attrToOid schema (Lcstring.of_string x)) toGenerate
