@@ -1,5 +1,8 @@
+open Ocamldap
 open Ooldap
+open Ooldif
 open Arg
+open Printf
 
 let _ = 
   (* stuff to handle command line args *)
@@ -24,12 +27,16 @@ let _ =
 
 
     (* do the ldap part *)
-    if (Array.length Sys.argv) > 9 then
-      (parse spec set_filter usg;
-       let ldap = new ldapcon !host ~port: !port in
-	 ldap#bind  !binddn ~cred: !cred;
-	 let print_entry e =
-	   e#print; print_endline "" in
-	   List.iter print_entry (ldap#search ~base: !base !filter))
-    else
-      usage spec usg
+    try
+      if (Array.length Sys.argv) > 9 then
+	(parse spec set_filter usg;
+	 let ldap = new ldapcon !host ~port: !port in
+	 let ldif = new ldif () in
+	   ldap#bind  !binddn ~cred: !cred;
+	   Ooldap.iter
+	     (fun e -> ldif#write_entry e) 
+	     (ldap#search_a ~base: !base !filter))
+      else
+	usage spec usg
+    with
+	LDAP_Failure e -> printf "ldap error: %s" (err2string e)
